package memcached

import (
	"log"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/common/tcp"
)

var IGNORE_METRICS = []string{
	"pid",
	"time",
	"version",
	"libevent",
	"pointer_size",
	"rusage_user",
	"rusage_system",
	"connection_structures",
	"reserved_fds",
	"hash_power_level",
	"lru_maintainer_juggles",
	"lrutail_reflocked",
	"lru_bumps_dropped",
}

var regexCacheCommand = regexp.MustCompile(`^(get|delete|incr|decr|cas|touch)_(hits|misses|badval|expired|flushed)$`)

func MakeDaemon(namespace string) *tcp.TcpDaemon {
	return &tcp.TcpDaemon{
		Timeout:        3 * time.Second, // TODO
		Namespace:      namespace,
		ServiceType:    "memcached",
		Keepalive:      true,
		CollectService: collect,
	}
}

func collect(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {
	collectStats(d, svc, m, target)
	collectSlabs(d, svc, m, target)
}

func collectStats(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {

	data, err := d.Command(svc, "stats\r\n", []string{"END", ""})
	if err != nil {
		m.IncrError("memcached", "memcached")
		log.Println("WARN", "in collecting memcached metrics", svc, err)
		return
	}

	for _, line := range data {
		kv := strings.SplitN(line, " ", 3)
		if len(kv) != 3 || kv[0] != "STAT" {
			// TODO log
			continue
		}
		stat := kv[1]
		valStr := kv[2]
		if common.Index(IGNORE_METRICS, stat) > -1 {
			continue
		}

		labels := [][]string{
			[]string{"namespace", d.Namespace},
			[]string{"service", svc.ID},
			[]string{"service_type", svc.Type},
			[]string{"service_addr", target},
		}

		val, errFloat := strconv.ParseFloat(valStr, 64)

		/* match commands get_hits, get_misses, delete_hits, ... */
		matches := regexCacheCommand.FindStringSubmatch(stat)
		if len(matches) == 3 {
			stat = "cache_commands_total"
			labels = append(labels, []string{"command", matches[1]})
			labels = append(labels, []string{"cache", matches[2]})

		} else {

			switch stat {
			case "max_connections":
				stat = "connections_max"
			case "curr_connections":
				stat = "connections"
			case "total_connections":
				stat = "connections_total"

			case "cmd_get":
				stat = "commands_total"
				labels = append(labels, []string{"command", "get"})
			case "cmd_set":
				stat = "commands_total"
				labels = append(labels, []string{"command", "set"})
			case "cmd_flush":
				stat = "commands_total"
				labels = append(labels, []string{"command", "flush"})
			case "cmd_touch":
				stat = "commands_total"
				labels = append(labels, []string{"command", "touch"})
			case "auth_cmds":
				stat = "commands_total"
				labels = append(labels, []string{"command", "auth"})
			case "auth_errors":
				stat = "commands_total"
				labels = append(labels, []string{"command", "auth_error"})

			case "bytes_read":
				stat = "io_bytes_total"
				labels = append(labels, []string{"direction", "in"})
			case "bytes_written":
				stat = "io_bytes_total"
				labels = append(labels, []string{"direction", "out"})

			case "listen_disabled_num":
				stat = "listen_disabled_total"
			case "time_in_listen_disabled_us":
				if errFloat != nil {
					continue
				}
				stat = "listen_disabled_duration_seconds"
				val /= 10000 // convert µs to seconds

			case "conn_yields":
				stat += "_total"

			case "slab_reassign_rescues":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "rescue"})
			case "slab_reassign_chunk_rescues":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "chunk_rescue"})
			case "slab_reassign_evictions_nomem":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "eviction_nomem"})
			case "slab_reassign_inline_reclaim":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "inline_reclaim"})
			case "slab_reassign_busy_items":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "busy_item"})
			case "slab_reassign_busy_deletes":
				stat = "slab_reassigns"
				labels = append(labels, []string{"type", "busy_delete"})

			case "slabs_moved":
				stat += "_total"

			case "log_worker_dropped":
				stat = "log_workers_total"
				labels = append(labels, []string{"state", "dropped"})
			case "log_worker_written":
				stat = "log_workers_total"
				labels = append(labels, []string{"state", "written"})
			case "log_watcher_skipped":
				stat = "log_watchers_total"
				labels = append(labels, []string{"state", "skipped"})
			case "log_watcher_sent":
				stat = "log_watchers_total"
				labels = append(labels, []string{"state", "sent"})

			case "bytes":
				stat = "size_bytes"
			case "limit_maxbytes":
				stat = "size_bytes_max"

			case "hash_bytes":
				stat = "hash_size_bytes"

			case "curr_items":
				stat = "stored_items"
			case "total_items":
				stat = "stored_items_total"

			case "lru_crawler_starts":
				stat += "_total"
			case "crawler_reclaimed":
				stat = "lru_" + stat + "_total"
			case "crawler_items_checked":
				stat = "lru_" + stat + "_total"

			case "moves_to_cold":
				stat = "lru_moves_total"
				labels = append(labels, []string{"type", "to_cold"})
			case "moves_to_warm":
				stat = "lru_moves_total"
				labels = append(labels, []string{"type", "to_warm"})
			case "moves_within_lru":
				stat = "lru_moves_total"
				labels = append(labels, []string{"type", "within_lru"})

			case "direct_reclaims":
				stat += "_total"

			case "uptime":
				stat += "_seconds"

				//			case "expired_unfetched":

			case "evictions":
				stat = "lru_evictions_total"
				labels = append(labels, []string{"type", "valid"})
			case "evicted_unfetched":
				stat = "lru_evictions_total"
				labels = append(labels, []string{"type", "unfetched"})
			case "evicted_active":
				stat = "lru_evictions_total"
				labels = append(labels, []string{"type", "active"})

			case "expired_unfetched":
				stat = "lru_expires_total"
				labels = append(labels, []string{"type", "unfetched"})

			case "reclaimed":
				stat = "lru_reclaimed_total"

			/* by default use the metrics as-is and consider it's a float */
			default:
				/* non exhaustive list:
				 *   accepting_conns
				 *   threads
				 *   hash_is_expanding
				 *   slab_reassign_running
				 *   lru_crawler_running
				 *   malloc_fails
				 *   slab_global_page_pool
				 *   rejected_connections
				 */
			}
		}
		if errFloat != nil {
			// TODO log
			continue
		}
		m.Set(&common.Metric{
			Name:   "memcached_" + stat,
			Labels: labels,
			Value:  val,
		})
	}
}

func collectSlabs(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {

	data, err := d.Command(svc, "stats slabs\r\n", []string{"END", ""})
	if err != nil {
		m.IncrError("memcached", "memcached")
		log.Println("WARN", "in collecting memcached metrics", svc, err)
		return
	}

	total_chunks := make(map[string]float64)
	chunk_size := make(map[string]float64)
	mem_requested := make(map[string]float64)

	for _, line := range data {
		kv := strings.SplitN(line, " ", 3)
		if len(kv) != 3 || kv[0] != "STAT" {
			// TODO log
			continue
		}
		stat := kv[1]
		valStr := kv[2]

		val, errFloat := strconv.ParseFloat(valStr, 64)
		if errFloat != nil {
			// TODO log
			continue
		}

		labels := [][]string{
			[]string{"namespace", d.Namespace},
			[]string{"service", svc.ID},
			[]string{"service_type", svc.Type},
			[]string{"service_addr", target},
		}

		if stat == "active_slabs" {
			m.Set(&common.Metric{
				Name:   "memcached_slab_actives_total",
				Labels: labels,
				Value:  val,
			})
			continue
		}

		if stat == "total_malloced" {
			m.Set(&common.Metric{
				Name:   "memcached_slab_memory_size_bytes_total",
				Labels: labels,
				Value:  val,
			})
			continue
		}

		matches := strings.SplitN(stat, ":", 2)
		if len(matches) != 2 {
			continue
		}
		slab := matches[0]
		stat = matches[1]
		labels = append(labels, []string{"slab", slab})

		switch stat {
		case "total_pages":
			stat = "pages_total"
		case "total_chunks":
			stat = "chunks_total"
			total_chunks[slab] = val
		case "chunk_size":
			stat += "_bytes"
			chunk_size[slab] = val
		case "mem_requested":
			stat += "_bytes"
			mem_requested[slab] = val

		case "get_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "get"})
		case "cmd_set":
			stat = "commands_total"
			labels = append(labels, []string{"command", "set"})
		case "delete_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "delete"})
		case "incr_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "incr"})
		case "decr_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "decr"})
		case "cas_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "cas"})
		case "cas_badval":
			stat = "commands_total"
			labels = append(labels, []string{"command", "cas_badval"})
		case "touch_hits":
			stat = "commands_total"
			labels = append(labels, []string{"command", "touch"})

		case "used_chunks":
			stat = "chunks_total"
			labels = append(labels, []string{"type", "used"})
		case "free_chunks":
			stat = "chunks_total"
			labels = append(labels, []string{"type", "free"})

		default:
			/*
			 *  chunks_per_page
			 *  free_chunks_end
			 */
		}

		m.Set(&common.Metric{
			Name:   "memcached_slab_" + stat,
			Labels: labels,
			Value:  val,
		})
	}

	/*
	 * from https://github.com/memcached/memcached/blob/master/doc/protocol.txt
	 * Items are stored in a slab that is the same size or larger than the
	 * item.  mem_requested shows the size of all items within a
	 * slab. (total_chunks * chunk_size) - mem_requested shows memory
	 * wasted in a slab class.  If you see a lot of waste, consider tuning
	 * the slab factor.
	 */
	for slab, total := range total_chunks {
		chunk, ok := chunk_size[slab]
		if !ok {
			continue
		}
		requested, ok := mem_requested[slab]
		if !ok {
			continue
		}
		m.Set(&common.Metric{
			Name: "memcached_slab_wasted_bytes",
			Labels: [][]string{
				[]string{"namespace", d.Namespace},
				[]string{"service", svc.ID},
				[]string{"service_type", svc.Type},
				[]string{"service_addr", target},
				[]string{"slab", slab},
			},
			Value: (total * chunk) - requested,
		})
	}
}
