package s3rt

import (
	"bytes"
	"context"
	"crypto/md5"
	"fmt"
	"log"
	"net/http"
	"strconv"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/awserr"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3"
	"github.com/aws/aws-sdk-go/service/s3/s3iface"
	"github.com/aws/aws-sdk-go/service/s3/s3manager"
)

type s3c struct {
	s3       s3iface.S3API
	bucketRT bool
	bucket   string
	service  *common.Service
}

type s3rtDaemon struct {
	daemon.Daemon

	clients       []*s3c
	bucketCreated []bool
	data          *[]byte
	errors        []*common.Service
	timeout       time.Duration
	namespace     string
}

const mpuSize = 5 * 1024 * 1024
const objSize = mpuSize + 1

func MakeDaemon(namespace string) *s3rtDaemon {
	data := make([]byte, objSize)

	return &s3rtDaemon{
		clients:   []*s3c{},
		data:      &data,
		errors:    []*common.Service{},
		timeout:   3 * time.Second, // TODO
		namespace: namespace,
	}
}

func (d *s3rtDaemon) ID() string {
	return "S3rt"
}

func (d *s3rtDaemon) Interval() time.Duration {
	return 10 * time.Second
}

func (d *s3rtDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *s3rtDaemon) OnServicesChange(services common.ServicesMap) error {
	d.reload(services)

	return nil
}

func (d *s3rtDaemon) reload(services common.ServicesMap) {
	swiftServices := services["oioswift"]

	clients := make([]*s3c, 0, len(swiftServices))
	errServices := make([]*common.Service, 0, len(swiftServices))
	for _, svc := range swiftServices {
		endpoint := fmt.Sprintf("http://%s:%d", svc.IP, svc.Port)
		config := aws.Config{
			Endpoint:         aws.String(endpoint),
			S3ForcePathStyle: aws.Bool(true),
			MaxRetries:       aws.Int(1),
			HTTPClient:       &http.Client{Timeout: d.timeout},
		}

		// Ensure to load config and credentials from ~/.aws
		sess, err := session.NewSessionWithOptions(session.Options{
			Config:            config,
			SharedConfigState: session.SharedConfigEnable,
		})
		if err != nil {
			log.Println("WARN", "in initializing client", svc, err)

			continue
		}
		clients = append(clients, &s3c{
			s3:       s3.New(sess),
			bucketRT: true,
			bucket:   fmt.Sprintf("s3rt-%x", md5.Sum([]byte(endpoint))),
			service:  svc,
		})
	}

	d.bucketCreated = make([]bool, len(clients))
	d.clients = clients
	d.errors = errServices
}

func (d *s3rtDaemon) Collect(m *common.MetricData) error {
	// TODO: do this only once?
	for _, svc := range d.errors {
		m.IncrError("s3rt", svc.Type)
	}

	var wg sync.WaitGroup
	for i, c := range d.clients {
		wg.Add(1)
		// Concurrent writing to different indices of a same slice is safe
		// ...but just to be sure
		go func(c *s3c, bucketCreated *bool) {
			defer wg.Done()

			var t time.Duration
			var err error

			if !*bucketCreated || c.bucketRT {
				t, err = c.mb(c.bucket)
				if err == nil {
					*bucketCreated = true
				} else if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyExists {
					*bucketCreated = true
				} else if !c.bucketRT {
					m.IncrError("s3rt", "s3rt")
					log.Println("WARN", "in precreating bucket", c, err)
				}
				if c.bucketRT {
					d.s3Publish(m, c.service, "mb", t, err, false)
				}
			}

			t, err = c.put(c.bucket, "s3roundtrip_object", d.data, mpuSize)
			d.s3Publish(m, c.service, "put", t, err, false)

			t, err = c.get(c.bucket, "s3roundtrip_object")
			d.s3Publish(m, c.service, "get", t, err, false)

			t, err = c.ls(c.bucket, 1000)
			d.s3Publish(m, c.service, "ls", t, err, false)

			t, err = c.del(c.bucket, "s3roundtrip_object")
			d.s3Publish(m, c.service, "del", t, err, false)

			// Time to first byte
			dataTtfb := []byte("0")
			t, err = c.put(c.bucket, "s3roundtrip_ttfb_object", &dataTtfb, mpuSize)
			d.s3Publish(m, c.service, "put", t, err, true)
			if err == nil {
				t, err = c.get(c.bucket, "s3roundtrip_ttfb_object")
				d.s3Publish(m, c.service, "get", t, err, true)

				_, _ = c.del(c.bucket, "s3roundtrip_ttfb_object")
			}

			if c.bucketRT {
				t, err = c.rb(c.bucket)
				d.s3Publish(m, c.service, "rb", t, err, false)
			}
		}(c, &d.bucketCreated[i])
	}

	wg.Wait()

	return nil
}

func (d *s3rtDaemon) s3Publish(m *common.MetricData, service *common.Service, request string, t time.Duration, err error, ttfb bool) {
	code := "2xx"
	if err != nil {
		if req, ok := err.(awserr.RequestFailure); ok {
			// Error is an AWS request failure
			status := req.StatusCode() / 100
			if status == 3 || status == 1 {
				// Treat redirects and continues as 2xx
				status = 2
			}
			code = strconv.Itoa(status) + "xx"
		} else {
			code = "other"
		}
	}
	labels := [][]string{
		[]string{"namespace", d.namespace},
		[]string{"service", service.ID},
		[]string{"service_type", service.Type},
		[]string{"service_addr", fmt.Sprintf("%s:%d", service.IP, service.Port)},
		[]string{"method", request},
		[]string{"code", code},
	}

	if ttfb {
		labels = append(labels, []string{"size", "0"})
	} else {
		labels = append(labels, []string{"size", strconv.Itoa(len(*d.data))})
	}
	m.Set(&common.Metric{
		Name:   "s3rt_response_time_seconds",
		Labels: labels,
		Value:  float64(t.Seconds()),
	})
}

func (c *s3c) put(bucket, object string, data *[]byte, mpuSize int64) (time.Duration, error) {
	uploader := s3manager.NewUploaderWithClient(c.s3)
	inputUpload := &s3manager.UploadInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(object),
		Body:   bytes.NewReader(*data),
	}
	ctx, cancel := context.WithTimeout(context.Background(), 30*time.Second)
	defer cancel()

	start := time.Now()
	_, err := uploader.UploadWithContext(ctx, inputUpload, func(u *s3manager.Uploader) {
		u.PartSize = mpuSize
	})
	return time.Since(start), err
}

func (c *s3c) get(bucket, object string) (time.Duration, error) {
	input := &s3.GetObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(object),
	}
	start := time.Now()
	_, err := c.s3.GetObject(input)
	return time.Since(start), err
}

func (c *s3c) del(bucket, obj string) (time.Duration, error) {
	input := &s3.DeleteObjectInput{
		Bucket: aws.String(bucket),
		Key:    aws.String(obj),
	}
	start := time.Now()
	_, err := c.s3.DeleteObject(input)
	return time.Since(start), err
}

func (c *s3c) mb(bucket string) (time.Duration, error) {
	input := s3.CreateBucketInput{
		Bucket: aws.String(bucket),
	}
	start := time.Now()
	_, err := c.s3.CreateBucket(&input)
	if aerr, ok := err.(awserr.Error); ok && aerr.Code() == s3.ErrCodeBucketAlreadyOwnedByYou {
		return time.Since(start), nil
	}
	return time.Since(start), err
}

func (c *s3c) rb(bucket string) (time.Duration, error) {
	input := s3.DeleteBucketInput{
		Bucket: aws.String(bucket),
	}
	start := time.Now()
	_, err := c.s3.DeleteBucket(&input)
	if err != nil {
		return time.Since(start), err
	}
	return time.Since(start), nil
}

func (c *s3c) ls(bucket string, keys int64) (time.Duration, error) {
	input := s3.ListObjectsInput{
		Bucket:  aws.String(bucket),
		MaxKeys: aws.Int64(keys),
	}
	start := time.Now()
	_, err := c.s3.ListObjects(&input)
	if err != nil {
		return time.Since(start), err
	}
	return time.Since(start), nil
}
